// *-*-c++-*-* for Emacs mode

#include <NTL/vec_vec_ZZ_p.h>
#include <NTL/ZZ_pX.h>
#include "vec_vec_ZZ_pX.h"
#include "ZZ_pX.h"

NTL_START_IMPL

/*
 * Plain evaluation/interpolation
 */

void PlainEval(ZZ_p& b, const ZZ_pX& f, const ZZ_p& a)
{
  eval(b, f, a);
}

void PlainEval(vec_ZZ_p& b, const ZZ_pX& f, const vec_ZZ_p& a)
{
  eval(b, f, a);
}

void TPlainEval(ZZ_pX& f, long d, const ZZ_p& b, const ZZ_p& a)
// transpose of Horner's evaluation b = f(a), where d=deg(f)
{
  long i;
  
  if (d < 0) {
    f.kill();
    return;
  }
  f.rep.SetLength(d+1);
  f.rep[0] = b;
  for (i = 1; i <= d; i++) {
    mul(f.rep[i], f.rep[i-1], a);
   }
  f.normalize();
}

void TPlainEval(ZZ_pX& f, long d, const vec_ZZ_p& b, const vec_ZZ_p& a)
// naive algorithm:  repeats transposed Horner
{
  long m = a.length();
  
  if (b.length() != m) Error("ZZ_pX1/TPlainEval: bad args");

  ZZ_pX g, h;
  long i;

  clear(h);
  for (i = 0; i < m; i++) {
    TPlainEval(g, d, b[i], a[i]);
    add(h, h, g);
  }
  f = h;
}

void PlainInterp(ZZ_pX& f, const vec_ZZ_p& a, const vec_ZZ_p& b)
{
  interpolate(f, a, b);
}

void TPlainInterp(vec_ZZ_p& b, const vec_ZZ_p& a, const ZZ_pX& f)
{
  long m = a.length();
  if (deg(f) >= m) Error("ZZ_pX1/TPlainInterp: vector length and degree mismatch");
  
  if (m == 0) {
    clear(b);
    return;
  }
  
  vec_vec_ZZ_p prod;
  prod.SetLength(m-1);
  prod[0].SetLength(1);
  negate(prod[0][0], a[0]);
  
  ZZ_p t1, t2, t3;
  
  long k, i;

  for (k = 1; k < m-1; k++) {
    prod[k].SetLength(k+1);
    const ZZ_p& aa = a[k];
    
    negate(t1, a[k]);
    add(prod[k][k], t1, prod[k-1][k-1]);
    for (i = k-1; i >= 1; i--) {
      mul(t2, prod[k-1][i], t1);
      add(prod[k][i], t2, prod[k-1][i-1]);
    }
    mul(prod[k][0], prod[k-1][0], t1);
  }
#ifdef DEBUG
  cout << "prod = " << prod << endl;
  cout << "a = " << a << endl;
#endif

  vec_ZZ_p res(INIT_SIZE,m);
  for (k = 0; k <= deg(f); k++) res[k]=f.rep[k];
  for (;k < m; k++) clear(res[k]);

  for (k = m-1; k >= 0; k--) {
    const ZZ_p& aa = a[k];
#ifdef DEBUG
    cout << "k = " << k << endl;
    cout << "res = " << res << endl;
    cout << "aa = " << aa << endl;
#endif
    t2 = res[k];
    for (i = 0; i < k; i++) {
      mul(t1, prod[k-1][i], res[i]);
      add(t2, t2, t1);
    }
#ifdef DEBUG
    cout << "t2 = " << t2 << endl;
#endif
    set(t1);
    for (i = k-1; i >= 0; i--) {
      mul(t1, t1, aa);
      add(t1, t1, prod[k-1][i]);
    }
#ifdef DEBUG
    cout << "t1 = " << t1 << endl;
#endif
    div(t2, t2, t1);
    res[k] = t2;

    for (i = 0; i < k; i++) {
      sub(res[i], res[i], t2);
      mul(t2, t2, aa);
    }
  }
  b = res;
}

/*
 * Classical fast evaluation, cf von zur Gathen & Gerhard's book
 * chapter 10.
 */

// NB: routines acting with trees are static and take care because
// they do not check memory clashes at all.

static
void BuildSubproductTree(vec_vec_ZZ_pX& tree, const vec_ZZ_pX& q, long depth)
  // Initialize the subproduct tree from q up to depth depth
{
  long n = q.length();
  long i, j, k, l, m, i2;

#ifdef DEBUG
  cout << "--> BuildSubproductTree: depth =" << depth << endl;
#endif

  tree.SetLength(depth);

  if (depth == 0) return;

  tree[0] = q;
  vec_ZZ_pX *p1 = &tree[0];
  vec_ZZ_pX *p2;

  l = n;
  for(j = 1; j < depth; j++) {
    p2 = &tree[j];
    k = l >> 1;
    m = l % 2;
    p2->SetLength(k + m);
    for(i = 0; i < k;  i++) {
      i2 = i << 1;
      mul((*p2)[i], (*p1)[i2], (*p1)[i2+1]);
    }
    if (m == 1)
      (*p2)[k] = (*p1)[k << 1];
    p1 = p2;
    l = k + m;
  }
#ifdef DEBUG
  cout << "<-- BuildSubproductTree: " << tree << endl;
#endif
}

static
void SubproductTreeDown(vec_ZZ_pX& r, const ZZ_pX& a, const vec_vec_ZZ_pX& tree)
{
  long d = tree.length();

  if (d == 0) {
    r.SetLength(1);
    r[0] = a;
    return;
  }

  r.SetLength(tree[0].length());

  long i, j, j2, h, n;
  const vec_ZZ_pX* p = &(tree[d-1]);

  for(j = 0; j < p->length(); j++) rem(r[j], a, (*p)[j]);

  for(i = d-2; i >= 0; i--) {
    p = &tree[i];
    n = p->length();
    h = n >> 1;
    if (n % 2 != 0) r[n-1] = r[h];
    for(j = h-1; j >=0; j--) {
      j2 = j << 1;
      rem(r[j2+1], r[j], (*p)[j2+1]);
      rem(r[j2], r[j], (*p)[j2]);
    }
  }
}

static
void TSubproductTreeDown(ZZ_pX& a, long da, const vec_ZZ_pX& r, const vec_vec_ZZ_pX& tree)
  // da = deg(a)
{
  long d = tree.length();

  clear(a);
  if (d == 0) {
    if (r.length() != 1) Error("ZZ_pX1/TSubproductTreeDown: bad args");
    a = r[0];
    return;
  }

  if (r.length() != tree[0].length()) Error("ZZ_pX1/TSubproductTreeDown: bad args");

  vec_ZZ_pX lr = r;
  ZZ_pX z;
  const vec_ZZ_pX* p;
  long i, j, j2, h, n, o;

  for(i = 0; i < d-1; i++) {
    p = &tree[i];
    n = p->length();
    h = n >> 1;

    for(j = 0; j < h; j++) {
      j2 = j << 1;
      o = deg(tree[i+1][j])-1;
      trem(z, o, (*p)[j2], lr[j2]);
      trem(lr[j], o, (*p)[j2+1], lr[j2+1]);
      add(lr[j], lr[j], z);
    }
    if (n % 2 != 0) lr[h] = lr[n-1];
  }
  p = &tree[d-1];
  for(j = 0; j < p->length(); j++) {
    trem(z, da, (*p)[j], lr[j]);
    add(a, a, z);
  }
}

static
void SubproductTreeUp(vec_ZZ_pX& a, const vec_ZZ_pX& r, const vec_vec_ZZ_pX& tree)
{
  long d = tree.length();

  if (d == 0) {
    a = r;
    return;
  }

  vec_ZZ_pX v = r;
  if (r.length() != tree[0].length()) Error("ZZ_pX1/SubproductTreeUp: bad args");
  long i, j, j2, h, n;
  const vec_ZZ_pX* p;
  ZZ_pX acc;
  for(i = 0; i < d; i++) {
    p = &tree[i];
    n = p->length();
    h = n >> 1;
    for(j = 0; j < h; j++) {
      j2 = j << 1;
      mul(acc, v[j2], (*p)[j2+1]);
      mul(v[j], v[j2+1], (*p)[j2]);
      add(v[j], v[j], acc);
    }
    if (n % 2 != 0) v[h] = v[n-1];
  }
  // TODO << use a in the last step of the loop to avoid copy
  a.SetLength(n);
  for(i = 0; i < n ; i++) a[i] = v[i];
}

static
void TSubproductTreeUp(vec_ZZ_pX& r, const vec_ZZ_pX& a, const vec_vec_ZZ_pX& tree)
{
#ifdef DEBUG
  cout << "a = " << a << endl;
#endif
  long d = tree.length();

  if (d == 0) {
    r = a;
    return;
  }
  if (a.length() != tree[d-1].length()) Error("ZZ_pX1/TSubproductTreeUp: length mismatch");

  long i, j, j2, h, n, o;
  const vec_ZZ_pX* p;

  r.SetLength(tree[0].length());
  for(i = 0; i < a.length(); i++) r[i] = a[i];

  for(i = d-2; i >= 0; i--) {
#ifdef DEBUG
    cout << "i = " << i << endl;
    cout << "tree[i] = " << tree[i] << endl;
    cout << "r = " << r << endl;
#endif
    p = &tree[i];
    n = p->length();
    h = n >> 1;
    if (n % 2 != 0) r[n-1] = r[h];
    for(j = h-1; j >= 0; j--) {
      j2 = j << 1;
      o = deg(tree[i+1][j])-1;
      tmul(r[j2+1], deg((*p)[j2+1]) - 1, (*p)[j2], r[j]);
      tmul(r[j2], deg((*p)[j2]) - 1, (*p)[j2+1], r[j]);
    }
  }
}

inline static
void BuildTwoByTwoProductsFromPoints(vec_ZZ_pX& v, const vec_ZZ_p& x)
  // Build the sequence (X-x[i])(X-x[i+1]) of consecutive products
  // NB: &x != &y
{
  long i, i2, hn, m, n = x.length();
  hn = n >> 1;
  m = n % 2;
  v.SetLength(hn + m);
  ZZ_p* T;
  for(i = 0; i < hn; i++) {
    i2 = i << 1;
    v[i].rep.SetLength(3);
    T = v[i].rep.elts();
    mul(T[0], x[i2], x[i2+1]);
    add(T[1], x[i2], x[i2+1]);
    negate(T[1],T[1]);
    T[2]=1;
  }
  if (m != 0) {
    v[hn].rep.SetLength(2);
    T = v[hn].rep.elts();
    negate(T[0], x[n-1]);
    T[1]=1;
  }
}

// Now we come to safe routines wrt memory
void Fast1Eval(vec_ZZ_p& y, const ZZ_pX& a, const vec_ZZ_p& x, const vec_vec_ZZ_pX& tree)
{
  long i, i2, j, k, hn, m, n = x.length();
  hn = n >> 1;
  m = n % 2;

  vec_ZZ_pX v;
  SubproductTreeDown(v, a, tree);
#ifdef DEBUG
  cout << "v = " << v << endl;
#endif

  y.SetLength(n);
  ZZ_p c0, c1;
  long dc;
  if (m != 0) {
    dc = deg(v[hn]);
    if (dc < 0) clear(y[n-1]);
    else y[n-1] = v[hn].rep[0];
  }
  for(i = hn-1; i >= 0; i--) {
    i2 = i << 1;
    dc = deg(v[i]);
    if (dc < 0) {
      clear(y[i2+1]);
      clear(y[i2]);
      continue;
    }
    c0 = v[i].rep[0];
    if (dc == 0) {
      y[i2+1] = c0;
      y[i2] = c0;
      continue;
    }
    c1 = v[i].rep[1];
    mul(y[i2+1], c1, x[i2+1]);
    add(y[i2+1], y[i2+1], c0);
    mul(y[i2], c1, x[i2]);
    add(y[i2], y[i2], c0);
  }
}

void Fast1Eval(vec_ZZ_p& y, const ZZ_pX& a, const vec_ZZ_p& x)
{
  long n = x.length();
  vec_ZZ_pX v;
  vec_vec_ZZ_pX tree;
  BuildTwoByTwoProductsFromPoints(v, x);
  BuildSubproductTree(tree, v, NextPowerOfTwo((n>>1) + (n%2)));
  Fast1Eval(y, a, x, tree);
}

void TFast1Eval(ZZ_pX& a, long da, const vec_ZZ_p& y, const vec_ZZ_p& x, const vec_vec_ZZ_pX& tree)
  // da = deg(a)
{
  long i, i2, j, k, hn, m, n = x.length();
  if (y.length()!=n) Error("ZZ_pX1/TFast1Eval: bad args");

  hn = n >> 1;
  m = n % 2;
  vec_ZZ_pX v(INIT_SIZE, hn + m);

  ZZ_p c;
  for(i = 0; i < hn; i++) {
    i2 = i << 1;
    v[i].rep.SetLength(2);
    add(v[i].rep[0], y[i2], y[i2 + 1]);
    mul(c, y[i2], x[i2]);
    mul(v[i].rep[1], y[i2+1], x[i2 + 1]);
    add(v[i].rep[1], v[i].rep[1], c);
  }
  if (m != 0) {
    v[hn].rep.SetLength(1);
    v[hn].rep[0] = y[n-1];
  }
#ifdef DEBUG
  cout << "v = " << v << endl;
#endif
  ZZ_pX tmp; // to avoid memory clash
  TSubproductTreeDown(tmp, da, v, tree);
  a = tmp;
}

void TFast1Eval(ZZ_pX& a, long da, const vec_ZZ_p& y, const vec_ZZ_p& x) {
  // da = deg(a)
  long n = x.length();
  if (y.length()!=n) Error("ZZ_pX1/TFast1Eval: bad args");

  vec_ZZ_pX v;
  vec_vec_ZZ_pX tree;
  BuildTwoByTwoProductsFromPoints(v, x);
  BuildSubproductTree(tree, v, NextPowerOfTwo((n>>1) + (n%2)));
  TFast1Eval(a, da, y, x, tree);
}

/*
 * New fast evaluation cf Bostan, Lecerf and Schost 2003.
 */

void Fast2Eval(vec_ZZ_p& y, const ZZ_pX& f, const vec_ZZ_p& x, const vec_vec_ZZ_pX& tree){

  long n = x.length();
  long hn = n >> 1;
  long m = n % 2;
  long i, i2;
  long d;
  ZZ_pX g, ig, res, test, P1, P3;

  d = tree.length();

  if (tree[d-1].length() !=1) Error("ZZ_pX1/Fast2Eval: bad tree");

  g = tree[d-1][0];

  CopyReverse(P1, f, 0, n-1);
  CopyReverse(ig, g, 0, n);
  NewtonInvTruncMul(P3, ig, n, P1);
  g=P3; // TODO << remove?
  trunc(g, g, n);

  vec_ZZ_pX T(INIT_SIZE, 1);
  T[0] = g;
  vec_ZZ_pX v;
  TSubproductTreeUp(v, T, tree);

  y.SetLength(n);
  ZZ_p *c;
  ZZ_p t;
  for(i = 0; i < hn; i++) {
    i2 = i << 1;
    if (deg(v[i]) < 0) {
      clear(y[i2]);
      clear(y[i2+1]);
      continue;
    }
    c = v[i].rep.elts();

    t = x[i2+1]; // &x may equal &y
    mul(y[i2+1], x[i2], 1);
    mul(y[i2+1], y[i2+1], c[0]);
    negate(y[i2+1], y[i2+1]);

    mul(y[i2], t, 1);
    mul(y[i2], y[i2], c[0]);
    negate(y[i2], y[i2]);

    if (deg(v[i]) == 0) continue;

    mul(t, c[1], 1);
    add(y[i2+1], y[i2+1], t);
    mul(t, c[1], 1);
    add(y[i2], y[i2], t);
    if (deg(v[i]) > 1) Error("Fast2Eval: bug");
  }
  if (m != 0) mul(y[n-1], coeff(v[hn],0), 1);

}

void Fast2Eval(vec_ZZ_p& y, const ZZ_pX& f, const vec_ZZ_p& x){

  long n = x.length();
  long hn = n >> 1;
  long m = n % 2;
  long i;
  long d;
  ZZ_pX g, ig, res, test;

  // Precomputations 
  vec_ZZ_pX w; 
  BuildTwoByTwoProductsFromPoints(w, x);
  vec_vec_ZZ_pX tree;
  BuildSubproductTree(tree, w, NextPowerOfTwo(hn + m) + 1);

  Fast2Eval(y, f, x, tree);
}

void TFast2Eval(ZZ_pX& f, long a, const vec_ZZ_p& y, const vec_ZZ_p& x, const vec_vec_ZZ_pX& tree)
{
  long n = x.length();
  long hn = n >> 1;
  long m = n % 2;
  long i, i2;
  long d;
  ZZ_pX g, ig, res, h;

  vec_ZZ_p z;
  z.SetLength(n);
  for(i = 0; i < n; i++) {
    z[i]=y[i];
  }

  vec_ZZ_pX v(INIT_SIZE, hn + m);
  ZZ_p *c;
  for(i = 0; i < hn; i++) {
    i2 = i << 1;
    v[i].rep.SetLength(2);
    c = v[i].rep.elts();
    mul(c[0], z[i2], x[i2+1]);
    mul(c[1], z[i2+1], x[i2]);
    add(c[0], c[1], c[0]);
    negate(c[0], c[0]);
    add(c[1], z[i2+1], z[i2]);
    if (IsZero(c[1])) {
      v[i].rep.SetLength(1);
      continue;
    }
  }
  if (m != 0) v[hn] = z[n-1];

  vec_ZZ_pX T;
  SubproductTreeUp(T, v, tree);

  d = tree.length();
  g = tree[d-1][0];

  CopyReverse(h,T[0],0,n-1);
  CopyReverse(ig,g,0,n);
  NewtonInvTruncMul(f, ig, n, h);

  trunc(f, f, n);
}

void TFast2Eval(ZZ_pX& f, long a, const vec_ZZ_p& y, const vec_ZZ_p& x)
{
  long n = x.length();
  long hn = n >> 1;
  long m = n % 2;
  long i;
  long d;
  ZZ_pX g, ig, res, test;

  // Precomputations 
  vec_ZZ_pX w; 
  BuildTwoByTwoProductsFromPoints(w, x);
  vec_vec_ZZ_pX tree;
  BuildSubproductTree(tree, w, NextPowerOfTwo(hn + m) + 1);

  TFast2Eval(f, a, y, x, tree);
}

namespace dvlp {

// TODO << ???
#define NTL_ZZ_pX_FAST1EVAL_CROSSOVER (32)
#define NTL_ZZ_pX_FAST2EVAL_CROSSOVER (400)

void eval(vec_ZZ_p& y, const ZZ_pX& a, const vec_ZZ_p& x, const vec_vec_ZZ_pX& tree)
{
  long m = x.length();
  if (m < NTL_ZZ_pX_FAST1EVAL_CROSSOVER)
    PlainEval(y, a, x);
  else if (m < NTL_ZZ_pX_FAST2EVAL_CROSSOVER)
    Fast1Eval(y, a, x, tree);
  else
    Fast2Eval(y, a, x, tree);
}

void eval(vec_ZZ_p& y, const ZZ_pX& a, const vec_ZZ_p& x)
{
  long m = x.length();
  if (m < NTL_ZZ_pX_FAST1EVAL_CROSSOVER)
    PlainEval(y, a, x);
  else if (m < NTL_ZZ_pX_FAST2EVAL_CROSSOVER)
    Fast1Eval(y, a, x);
  else
    Fast2Eval(y, a, x);
}

void teval(ZZ_pX& a, long da, const vec_ZZ_p& y, const vec_ZZ_p& x, const vec_vec_ZZ_pX& tree)
{
  long m = x.length();
  if (m < NTL_ZZ_pX_FAST1EVAL_CROSSOVER)
    TPlainEval(a, da, y, x);
  else if (m < NTL_ZZ_pX_FAST2EVAL_CROSSOVER)
    TFast1Eval(a, da, y, x, tree);
  else
    TFast2Eval(a, da, y, x, tree);
}

void teval(ZZ_pX& a, long da, const vec_ZZ_p& y, const vec_ZZ_p& x)
{
  long m = x.length();
  if (m < NTL_ZZ_pX_FAST1EVAL_CROSSOVER)
    TPlainEval(a, da, y, x);
  else if (m < NTL_ZZ_pX_FAST2EVAL_CROSSOVER)
    TFast1Eval(a, da, y, x);
  else
    TFast2Eval(a, da, y, x);
}

} // end ns dvlp

/*
 * Interpolation, cf von zur Gathen & Gerhard's book chapter 10.
 */

void FastInterp(ZZ_pX& f, const vec_ZZ_p& x, const vec_ZZ_p& y, const vec_vec_ZZ_pX& tree)
  // f(x[i]) = y[i]
  // if the x[i] are not pairwise distinct a division by 0 will be raised.
{
  long i, i2, d, hn, m, n = x.length();
  hn = n >> 1;
  m = n % 2;

  if (y.length() != n) Error("ZZ_pX1/FastInterp: vector length mismatch");

  d = tree.length();
  ZZ_pX g;
  if (tree[d-1].length() !=1) Error("ZZ_pX1/FastInterp: bad tree");

  diff(g, tree[d-1][0]);

  vec_ZZ_p z;
  dvlp::eval(z, g, x, tree);
  for(i = 0; i < n; i++) {
    div(z[i], y[i], z[i]);
  }

  vec_ZZ_pX v(INIT_SIZE, hn + m);
  ZZ_p *c;
  for(i = 0; i < hn; i++) {
    i2 = i << 1;
    v[i].rep.SetLength(2);
    c = v[i].rep.elts();
    mul(c[0], z[i2], x[i2+1]);
    mul(c[1], z[i2+1], x[i2]);
    add(c[0], c[0], c[1]);
    negate(c[0], c[0]);
    add(c[1], z[i2+1], z[i2]);
    if (IsZero(c[1])) {
      v[i].rep.SetLength(1);
      continue;
    }
  }
  if (m != 0) v[hn] = z[n-1];
#ifdef DEBUG
  cout << "v = " << v << endl;
#endif
  vec_ZZ_pX T;
  SubproductTreeUp(T, v, tree);
  f = T[0];
}

void FastInterp(ZZ_pX& f, const vec_ZZ_p& x, const vec_ZZ_p& y) {
  // f(x[i]) = y[i]
  long i, i2, d, hn, m, n = x.length();
  hn = n >> 1;
  m = n % 2;

  if (y.length() != n) Error("ZZ_pX1/FastInterp: vector length mismatch");

  vec_ZZ_pX v;
  BuildTwoByTwoProductsFromPoints(v, x);
  vec_vec_ZZ_pX tree;
  BuildSubproductTree(tree, v, NextPowerOfTwo(hn + m) + 1);
  FastInterp(f, x, y, tree);
}

void TFastInterp(vec_ZZ_p& y, const vec_ZZ_p& x, const ZZ_pX& f, const vec_vec_ZZ_pX& tree)
  // transposed of f(x[i]) = y[i]
  // if the x[i] are not pairwise distinct a division by 0 will be raised
{
  long i, i2, d, hn, m, n = x.length();
  hn = n >> 1;
  m = n % 2;

  if (deg(f) > n-1) Error("ZZ_pX1/TFastInterp: vector length mismatch");

  d = tree.length();
  ZZ_pX g;
  if (tree[d-1].length() !=1) Error("ZZ_pX1/TFastInterp: bad tree");
  diff(g, tree[d-1][0]);
  vec_ZZ_p z;
  dvlp::eval(z, g, x, tree);
#ifdef DEBUG
  cout << "z = " << z << endl;
#endif

  vec_ZZ_pX T(INIT_SIZE, 1);
  T[0] = f;
  vec_ZZ_pX v;
  TSubproductTreeUp(v, T, tree);
#ifdef DEBUG
  cout << "v = " << v << endl;
#endif

  for(i = 0; i < n; i++) {
    inv(z[i], z[i]);
  }
#ifdef DEBUG
  cout << "z = " << z << endl;
#endif

  y.SetLength(n);
  ZZ_p *c;
  ZZ_p t;
  for(i = 0; i < hn; i++) {
    i2 = i << 1;
    if (deg(v[i]) < 0) {
      clear(y[i2]);
      clear(y[i2+1]);
      continue;
    }
    c = v[i].rep.elts();

    t = x[i2+1]; // &x may equal &y
    mul(y[i2+1], x[i2], z[i2+1]);
    mul(y[i2+1], y[i2+1], c[0]);
    negate(y[i2+1], y[i2+1]);

    mul(y[i2], t, z[i2]);
    mul(y[i2], y[i2], c[0]);
    negate(y[i2], y[i2]);

    if (deg(v[i]) == 0) continue;

    mul(t, c[1], z[i2+1]);
    add(y[i2+1], y[i2+1], t);
    mul(t, c[1], z[i2]);
    add(y[i2], y[i2], t);
#ifdef DEBUG
    cout << "y = " << y << endl;
#endif
    if (deg(v[i]) > 1) Error("TFastInterp: bug");
  }
  if (m != 0) mul(y[n-1], coeff(v[hn],0), z[n-1]);
}

void TFastInterp(vec_ZZ_p& y, const vec_ZZ_p& x, const ZZ_pX& f) {
  // f(x[i]) = y[i]
  long i, i2, d, hn, m, n = x.length();
  hn = n >> 1;
  m = n % 2;

  if (deg(f) > n-1) Error("ZZ_pX1/TFastInterp: vector length mismatch");

  vec_ZZ_pX v;
  BuildTwoByTwoProductsFromPoints(v, x);
  vec_vec_ZZ_pX tree;
  BuildSubproductTree(tree, v, NextPowerOfTwo(hn + m) + 1);
  TFastInterp(y, x, f, tree);
}


NTL_END_IMPL

